# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

import numpy as np
import qutip


def test_unitary_evolution_two_level_system():
    delta = 1.0 * 2 * np.pi
    eps0 = 1.0 * 2 * np.pi
    A = 0.5 * 2 * np.pi
    omega = np.sqrt(delta**2 + eps0**2)
    T = 2*np.pi / omega
    tlist = np.linspace(0, 2*T, 101)
    psi0 = qutip.rand_ket(2)
    H0 = -0.5*eps0*qutip.sigmaz() - 0.5*delta*qutip.sigmax()
    H1 = 0.5 * A * qutip.sigmax()
    args = {'w': omega}
    H = [H0, [H1, lambda t, args: np.sin(args['w'] * t)]]
    e_ops = [qutip.num(2)]

    trial = qutip.fsesolve(H, psi0, tlist, e_ops, T, args).expect[0]
    expected = qutip.mesolve(H, psi0, tlist, [], e_ops, args).expect[0]

    np.testing.assert_allclose(trial, expected, atol=1e-4)
